@@ -435,6 +435,22 @@ void optimize(Env& env, bool chain_compression) {
               top_down(cd, c);
               ci->e(envi.constants.literalTrue);
               ci->remove();
+            } else if ((c->id() == envi.constants.ids.int_.eq ||
+                        c->id() == envi.constants.ids.bool_.eq ||
+                        c->id() == envi.constants.ids.float_.eq ||
+                        c->id() == envi.constants.ids.set_.eq) &&
+                       ((Expression::isa<Id>(c->arg(0)) &&
+                         Expression::cast<Id>(c->arg(0))->decl()->e() == nullptr &&
+                         Expression::type(c->arg(1)).isPar()) ||
+                        (Expression::isa<Id>(c->arg(1)) &&
+                         Expression::cast<Id>(c->arg(1))->decl()->e() == nullptr &&
+                         Expression::type(c->arg(0)).isPar()))) {
+              // equality constraint with one fixed var can be resolved later by
+              // simplify_bool_constraint
+              auto* id = Expression::cast<Id>(c->arg(Expression::isa<Id>(c->arg(0)) ? 0 : 1));
+              int idx = envi.varOccurrences.find(id->decl());
+              push_vardecl(envi, m[idx]->cast<VarDeclI>(), idx, vardeclQueue);
+              push_dependent_constraints(envi, id, constraintQueue);
             } else if (c->id() == envi.constants.ids.int_.lin_eq &&
                        Expression::equal(c->arg(2), IntLit::a(0))) {
               auto* al_c = Expression::cast<ArrayLit>(follow_id(c->arg(0)));
@@ -834,6 +850,16 @@ void optimize(Env& env, bool chain_compression) {
           }
         } else if (vd->type().isint() && (vd->ti()->domain() != nullptr)) {
           IntSetVal* isv = eval_intset(envi, vd->ti()->domain());
+          if (auto* il = Expression::dynamicCast<IntLit>(vd->e())) {
+            auto iv = IntLit::v(il);
+            if (!isv->contains(iv)) {
+              env.envi().fail();
+            } else if (isv->size() != 1 || isv->card() != 1) {
+              isv = IntSetVal::a(iv, iv);
+              vd->ti()->domain(new SetLit(Location().introduce(), isv));
+              push_dependent_constraints(envi, vd->id(), constraintQueue);
+            }
+          }
           if (isv->size() == 1 && isv->card() == 1) {
             simplify_constraint(envi, m[var_idx], deletedVarDecls, constraintQueue, vardeclQueue);
           }
@@ -1258,7 +1284,12 @@ bool simplify_constraint(EnvI& env, Item* ii, std::vector<VarDecl*>& deletedVarD
         push_dependent_constraints(env, Expression::cast<Id>(c->arg(0)), constraintQueue);
         CollectDecls cd(env, env.varOccurrences, deletedVarDecls, ii);
         top_down(cd, c);
-        ii->remove();
+        if (auto* vdi = ii->dynamicCast<VarDeclI>()) {
+          vdi->e()->e(env.constants.literalTrue);
+          deletedVarDecls.push_back(vdi->e());
+        } else {
+          ii->remove();
+        }
       } else if (Expression::type(c->arg(0)).isPar() && Expression::type(c->arg(1)).isPar()) {
         Expression* e0 = eval_par(env, c->arg(0));
         Expression* e1 = eval_par(env, c->arg(1));
